
from argparse import ArgumentParser
import numpy as np
import pandas as pd
import nibabel as nb

from ssm_pca.image_preprocess import preprocess
from ssm_pca.masking import mask_img
from notebooks.get_rois import get_rois
from ssm_pca.selecting_best_pc import normalize_vectors


def tpr(filepath, mask_file, gis_pc, group_mean_profile, to_mni_space=True):
    """
    Returns:
        object:
    """
    img = nb.load(filepath)
    array = img.get_fdata()

    array = preprocess(array, to_mni_space=to_mni_space)
    
    mask = nb.load(mask_file).get_fdata()

    # Get mask in row format
    mask_row = mask.flatten()

    # Mask with given file
    #array = mask_img(array, mask_file=mask_file)
    array = array * mask

    vector = array.flatten()

    # Change the mask so that 0 value of subject is removed   
    mask_row = mask_row*(vector > 0)
    mask_sum = np.sum(mask_row)

    # Replace all values <= 0 with 1 (to be able to do the log)
    vector[vector <= 0] = 1

    # Log transform vector
    log_vector = np.log(vector)

    # Subtract mean of subject
    subject_mean = np.sum(log_vector)/mask_sum
    centered_vector = log_vector - subject_mean

    # Mask again
    centered_vector = mask_row * centered_vector

    # Compute subject residual profiles (SRP) - subtract group mean profile
    srp = centered_vector - group_mean_profile

    # Mask again
    srp = mask_row * srp

    gis_pc = normalize_vectors(gis_pc)

    score = np.dot(srp, gis_pc)

    return srp, score

def tpr_roi(filepath, mask_file, roi_mask_file, gis_pc, group_mean_profile, to_mni_space=True):
    """

    Returns:
        object:
    """
    img = nb.load(filepath)
    array = img.get_fdata()

    if to_mni_space:
        array = np.pad(array, ((6, 6), (7, 7), (11, 11)))

    array = mask_img(array, mask_file=mask_file)

    roi_mask = nb.load(roi_mask_file).get_fdata()

    array, labels = get_rois(array, rois_mask=roi_mask)

    array = array / np.mean(array)

    # Log transform group matrix
    array = np.log(array)

    subject_mean = np.mean(array)

    vector = array

    srp = vector - subject_mean - group_mean_profile

    score = np.dot(srp, gis_pc)

    return srp, score


def prospective_scores(filelist, pc, mask_file, gis_pc_file, group_mean_profile_file, csv_scores, labels=None, to_mni_space=True, save_dir=None):

    gis_pc = np.load(gis_pc_file)
    gmp = np.load(group_mean_profile_file)
    df_scores_deriv = pd.read_csv(csv_scores)

    scores = []

    for filepath in filelist:
        _, score = tpr(filepath, mask_file, gis_pc, gmp, to_mni_space)
        scores.append(score)

    # Read derivation scores and Z-transform them 
    df_pc = df_scores_deriv[df_scores_deriv['PC'] == pc]

    scores_der = df_pc['scores'].to_numpy()
    labels_der = df_pc['labels'].to_numpy()

    scores_der_z = (np.array(scores_der) - np.mean(scores_der, axis=0)) / np.std(scores_der, axis=0, ddof=1)
    controls_mean = np.mean(scores_der_z[labels_der == 0])

    # Use the same mean, std and controls mean to Z-transform the prospective scores
    scores_z = (np.array(scores) - np.mean(scores_der, axis=0)) / np.std(scores_der, axis=0, ddof=1)
    scores_z = scores_z - controls_mean
    scores_z = scores_z.flatten()
    #print(scores_z)
    #print(scores_z.flatten())

    if save_dir:
        # Save the scores into a dataframe
        if labels:
            data = {'filepaths': filelist, 'labels': labels, 'scores': scores_z, 'PC': pc}
        else:
            data = {'filepaths': filelist, 'scores': scores_z, 'PC': pc}
            
        df = pd.DataFrame(data)
        df.to_csv(save_dir + f'/scores_prospective_pc_{pc}.csv')

    return scores_z, df


if __name__ == '__main__':
    #Parse config file path

    parser = ArgumentParser()
    parser.add_argument("--csv_file", type=str, help='CSV file with filepaths and corresponding labels of normal controls (0) or diseased subjects (1). Columns should be filepaths and labels.')
    parser.add_argument("--mask_file", type=str, help='Path to a binary brain mask file to apply to the images.')
    parser.add_argument("--gis_pc_file", type=str, help='Path to the GIS pattern (.npy) you want to get the scores from.')
    parser.add_argument("--gmp", type=str, help='Path to the group mean profile file (.npy) of discovery set.')
    parser.add_argument("--csv_scores", type=str, help='Path to the scores of discovery set.')
    parser.add_argument("--save_dir", type=str, help='Directory path where to save the results files.')
    
    args = parser.parse_args()
    
    df = pd.read_csv(args.csv_file)

    filelist = list(df['filepaths'])


    prospective_scores(filelist, mask_file=args.mask_file, gis_pc_file=args.gis_pc_file, group_mean_profile_file=args.gmp, 
                      csv_scores=args.csv_scores, to_mni_space=True, save_dir=args.save_dir)
 



import numpy as np
import nibabel as nib


def preprocess(array, to_mni_space=True, ref_region_mask_file='/Users/leonorlopes/Documents/data/masks/SPM_masks/brainmask.nii'):
    """Function to normalize the images by dividing by the whole brain uptake."""

    if to_mni_space:
        array = np.pad(array, ((6, 6), (7, 7), (11, 11)))

    array = np.float64(array)
    mask = nib.load(ref_region_mask_file).get_fdata()

    # Mask array
    masked = array * mask

    # Select only non-zero values of array and flatt it
    masked = np.ndarray.flatten(masked)
    masked = masked[np.where(masked != 0)]

    # Get mean of ref region
    mean_ref = np.mean(masked)

    array = array / mean_ref

    return array



# from .image_preprocess import *
# from .masking import *
# from .pca_analysis import *
# from .voxel_pattern_analysis import *
# from .visualization import *
# from .prospective_score import *
# from .selecting_best_pc import *
# from .main_old import *
# from .main_roi import *
# from .match_age_sex import *





from argparse import ArgumentParser
import os
import numpy as np
import nibabel as nb
import pandas as pd
import scipy.linalg as la
import matplotlib.pyplot as plt

from ssm_pca.image_preprocess import preprocess
from ssm_pca.masking import create_grey_matter_mask
from ssm_pca.pca import apply_pca, plot_eigenvalues_vaf
from ssm_pca.utils import pc_scores_to_df, display_gis
from ssm_pca.selecting_best_pc import best_pc_combination, combine_pc_vectors, compare_scores_normal_disease


def ssm_pca(filelist=None, labels=None, preprocess_img=True, mask_file=None, save_dir=None):

    """ Main function to run Scaled Subprofile Model with Principal Component Analysis on the images on filelist.

    Args:
        filelist: List with paths to the images to use. Defaults to None.
        labels: List to the labels corresponding to the images. Defaults to None.
        preprocess_img: Whether to preprocess the images. Defaults to True.
        mask_file: Path to a mask file to mask the images or threshold to create mask. If None, no mask is applied Defaults to None.
        save_dir: Directory to save the results. If None, results are not saved. Defaults to None.

    Returns:
        z_transf_gis - z_transformed_patterns: 2D numpy array of Z-transformed voxel pattern vectors (n_voxels, n_components)  
        gis - voxel pattern eigenvectors
        vaf - variace accounted for by each component
        gmp - group mean profile
    """

    
    n_subj = len(filelist)
    print(f'Number of subjects: {n_subj}')
    n_d = np.count_nonzero(labels)
    print(f'Number of diseased subjects: {n_d}')
    n_nc = n_subj - n_d
    print(f'Number of normal controls: {n_nc}')

    # Create save directories
    vector_save_dir = save_dir + '/vectors'
    os.makedirs(vector_save_dir, exist_ok=True)

    nifti_save_dir = save_dir + '/nifti'
    os.makedirs(nifti_save_dir, exist_ok=True)

    plots_save_dir = save_dir + '/plots'
    os.makedirs(plots_save_dir, exist_ok=True)


    # Load mask
    if mask_file:
        if type(mask_file) == float:
            mask = create_grey_matter_mask(filelist, threshold=float(mask_file), save_name=save_dir + '/threshold_mask.nii')
        else:
            mask = nb.load(mask_file).get_fdata()

    row_vectors_list = []

    # Load images and create row vectors
    for filepath in filelist:
        img = nb.load(filepath)
        array = img.get_fdata()

        # Preprocess image (change dimensions, spatial normalize, intensity normalize, etc)
        if preprocess_img:
            array = preprocess(array)

        # Mask with loaded mask
        if mask_file:
            array = array * mask

        # Convert the array to a row vector
        row_vector = array.flatten()

        row_vectors_list.append(row_vector)

    print('Creating group matrix...')
    # Create group matrix
    subj_matrix = np.vstack(row_vectors_list)

    # Get mask in row format for later use
    mask_row = mask.flatten()
    mask_sum = np.sum(mask_row) # Number of voxels in the mask

    # Replace all values <= 0 with 1 (to be able to do the log)
    subj_matrix[subj_matrix <= 0] = 1

    print('Log transforming data...')
    # Log transform the data
    log_data_matrix = np.log(subj_matrix)

    # Center the data matrix (subtract the row means). Get GMP - column means.
    subject_means = np.sum(log_data_matrix, axis=1)/mask_sum

    centered_data_matrix = log_data_matrix - subject_means[:, np.newaxis]

    # Mask again
    centered_data_matrix = centered_data_matrix * mask_row
  
    print('Computing SRP...')
    # Compute subject residual profiles (SRP) - subtract column means
    voxels_means = np.mean(centered_data_matrix, axis=0, keepdims=True)
    subject_residual_profiles = centered_data_matrix - voxels_means 

    if save_dir:
        np.save(vector_save_dir + '/group_mean_profile.npy', voxels_means)
  
    # Mask again 
    subject_residual_profiles = subject_residual_profiles * mask_row

    # Compute subject-by-subject covariance matrix
    covariance_matrix = subject_residual_profiles @ subject_residual_profiles.T

    print('Appliying PCA...')
    # Apply PCA to the covariance matrix
    score_vectors, eigVal, vaf = apply_pca(covariance_matrix)

    if save_dir:
        np.save(vector_save_dir + f'/score_vectors.npy', score_vectors)

    if save_dir:
        # Plot the variance accounted for (vaf) by each component
        plot_eigenvalues_vaf(eigVal, vaf)    

    # GIS vector and its negative form are mathematically equivalent solutions of the eigenvector equation 
    # with a corresponding sign flip of associated subject scores. 
    # Only the GIS form corresponding to positive mean patient scores 
    # relative to normal subject scores is considered to be relevant.

    print('Changing sign of scores if needed...')
    # Flip signs of SSFs if needed
    if n_nc > 0:
        for i in range(len(eigVal)):
            mean_nc = np.mean(score_vectors[:n_nc, i])
            mean_disease = np.mean(score_vectors[n_nc:, i])
            if mean_nc > mean_disease:
                print(f'change sign of PC: {i}')
                score_vectors[:, i] *= -1

    print('Computing GIS...')
    # Compute voxel pattern eigenvectors or GIS
    gis = subject_residual_profiles.T @ score_vectors

    print('Z-transforming GIS...')
    # Z-transform the GIS
    for i in range(n_subj):
        mean = np.sum(gis[:, i]) / mask_sum
        std = np.std(gis[gis[:, i] != 0, i])
        gis[:, i] = (gis[:, i] - mean) / std

    # Mask the GIS
    gis = gis * mask_row[:, np.newaxis]

    if save_dir:
        np.save(vector_save_dir + f'/GIS.npy', gis)

    return gis, score_vectors, vaf, voxels_means


def pattern_biomarker_analysis(gis, score_vectors, vaf, filelist, labels, img_shape=(91,109,91), save_dir=None):
    
    all_score_dfs = []

    print('Analysing PC scores (NC and disease comparisons)...')
    for pc, vaf_pc in enumerate(vaf):
        if vaf_pc > 5:
            display_gis(gis, pc, img_shape, save_dir) # Display GIS in image format
            df = pc_scores_to_df(score_vectors, pc, filelist, labels) #Put current PC scores into a dataframe
            p = compare_scores_normal_disease(df, pc, labels, save_dir) #Plot comparisons between normal controls and diseased in this PC
            all_score_dfs.append(df)
    
    df_all_scores = pd.concat(all_score_dfs)

    all_pcs = list(df_all_scores['PC'].unique())

    print('Comparing all PC combinations...')
    # Logistic regression to find the best combination of PCs
    df_aic, df_vafs, df_p, df_all_scores, coefs = best_pc_combination(df_all_scores, vaf, save_dir)


    print('Combining PC vectors...')
    results = combine_pc_vectors(gis, coefs, save_dir)

    df_all_scores.to_csv(save_dir + '/scores.csv')

    
    if save_dir:
        df_aic.to_csv(save_dir + '/AIC.csv')
        df_p.to_csv(save_dir + '/p_values.csv')
        df_vafs.to_csv(save_dir + '/VAFs.csv')

    print('AIC:\n')
    print(df_aic.to_string())
    
    print('Vaf:\n')
    print(df_vafs.to_string())

    print('P-values:\n')
    print(df_p.to_string())




if __name__ == '__main__':

    # Parse config file path
    parser = ArgumentParser()
    parser.add_argument("--csv_file", type=str, help='CSV file with filepaths and corresponding labels of normal controls (0) or diseased subjects (1). Columns should be filepaths and labels.')
    parser.add_argument("--mask_file", help='Path to a binary brain mask file to apply to the images or threshold to create mask.')
    parser.add_argument("--image_shape", type=tuple, help='Image shape to reconstruct back the PC pattern images. Default is the MNI brain space dimensions.',
                        default=(91, 109, 91))
    parser.add_argument("--preprocess_img", type=bool, help='Whether to preprocess the image.')
    parser.add_argument("--save_dir", type=str, help='Directory path where to save the results files.')
    
    args = parser.parse_args()
    
    df = pd.read_csv(args.csv_file)

    filelist = list(df['filepaths'])
    labels = list(df['labels'])


    gis, score_vectors, vaf, voxels_means = ssm_pca(filelist, labels,
                                    preprocess_img=args.preprocess_img,
                                    mask_file=args.mask_file,
                                    save_dir=args.save_dir)

    pattern_biomarker_analysis(gis, score_vectors, vaf, filelist, labels, img_shape=args.image_shape, 
                            save_dir=args.save_dir)




import numpy as np
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
import scipy.linalg as la

def apply_pca(covariance_matrix):
    """
    Apply PCA to the subject-by-subject covariance matrix C.

    Args:
        covariance_matrix: Subject-by-subject covariance matrix C.

    Returns:
        score_vectors: Projections of the original data points into the eigenvector space.
        eigenvalues: Eigenvalues of the covariance matrix.
        eigenvectors: Eigenvectors of the covariance matrix.
    """

    eigVal, eigVec = la.eigh(covariance_matrix)

    eigVal = np.real(eigVal)
    eigVec = np.real(eigVec)
    
    # Calculate the variance accounted for (vaf) by each component
    vaf = (eigVal / np.sum(eigVal))*100

    # Weight each score vector by the square root of its corresponding eigenvalue
    sqrt_eigVal = np.sqrt(eigVal)
    sqrt_eigVal = np.tile(sqrt_eigVal, (eigVec.shape[0], 1))

    score_vectors = eigVec * sqrt_eigVal 

    # score_vectors is a matrix of (subjects x GIS)
    # Each column represents a weighted eigenvector (dominant pattern of variability across subjects). 
    # Each element of the vector represents how the subject contributes or aligns with this pattern.
  
    # Reverse the order of eigenvalues, eigenvectors, and percent variance
    score_vectors = np.fliplr(score_vectors)
    vaf = np.flip(vaf)  # Percent variance accounted for
    eigVal = np.flip(eigVal)  # Eigenvalues


    return score_vectors, eigVal, vaf

def plot_eigenvalues_vaf(eigenvalues, vaf):

    fig, (ax1, ax2) = plt.subplots(2, 1)

    components = np.arange(1, len(vaf) + 1)

    ax1.plot(components, eigenvalues, color='skyblue', marker='o')
    #ax1.set_xlabel('Principal Component')
    ax1.set_ylabel('Eigenvalue')
    ax1.set_xlim([0,20])

    # Disable x-axis ticks for the top plot
    ax1.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=False)

    ax2.plot(components, vaf, color='skyblue', marker='o')
    ax2.set_xlabel('Principal Component')
    ax2.set_ylabel('% Vaf')
    ax2.set_xlim([0, 20])
    ax2.set_xticks(components)

    fig.suptitle('Eigenvalues and % Vaf by Each Principal Component')
    #plt.xticks(components)

    plt.grid(False)
    plt.show(block=False)





import pandas as pd

def match_age_sex_range(df_patients, df_controls, age_range=5, age_col='age', sex_col='sex'):

    matched_controls = []
    # Convert age columns to numeric type
    df_patients[age_col] = df_patients[age_col].astype(int)
    df_controls[age_col] = df_controls[age_col].astype(int)

    # Create a copy of df_controls to keep track of available controls
    available_controls = df_controls.copy()

    available_controls = available_controls.reset_index()
    df_patients = df_patients.reset_index()

    # Iterate through each PD patient
    for index, patient in df_patients.iterrows():

        # Define the age range for matching
        min_age = patient[age_col] - age_range
        max_age = patient[age_col] + age_range

        try:
            # Find controls with matching sex and age within the defined range
            matched_control = available_controls[(available_controls[sex_col] == patient[sex_col]) &
                                          (available_controls[age_col] >= min_age) &
                                          (available_controls[age_col] <= max_age)].sample(n=1, random_state=40) #42 not bad

            # If a matching control is found, add it to the list
            if not matched_control.empty:
                matched_controls.append(matched_control)
                # Remove the matched control from the available_controls DataFrame
                available_controls = available_controls.drop(matched_control.index)
            else:
                print(f'No matching control found for patient {index} with age {patient[age_col]}')

        except ValueError as e:
            print(e)

    # Concatenate the matched controls into a single DataFrame
    matched_controls_df = pd.concat(matched_controls)
    return matched_controls_df



from itertools import permutations, combinations
import numpy as np
import os
import matplotlib.pyplot as plt
import pandas as pd
from scipy.stats import ttest_ind
from sklearn.metrics import roc_auc_score, RocCurveDisplay, log_loss
from sklearn.linear_model import LogisticRegression

from ssm_pca.utils import reshape_eigenvector_to_3d, reshape_eigenvector_to_roi
from ssm_pca.visualization import plot_comparisons


def best_pc_combination(df, vaf=None, save_dir=None):

    df_temp = df.copy()

    pcs = list(df_temp['PC'].unique())

    # Get all possible combinations
    all_combinations = []
    for r in range(1, len(pcs) + 1):
        all_combinations.extend(combinations(range(len(pcs)), r))

    aics = dict()
    vafs_dict = dict()
    p_dict = dict()
    coefs_dict = dict()
    resulting_scores_dict = dict()

    labels = list(df_temp[df_temp['PC'] == 0]['labels'])
    filepaths = list(df_temp[df_temp['PC'] == 0]['filepaths'])

    # Iterate over the different combinations
    for combination in all_combinations:
        scores_comb = []
        vafs = []
    
        for i in combination:
            df_i = df_temp[df_temp['PC'] == i]
            scores_i = list(df_i['scores'])
            scores_comb.append(scores_i)
            vafs.append(vaf[i])

        scores = np.array(scores_comb).T

        # Fit logistic regression model
        logistic_model = LogisticRegression()
        logistic_model.fit(scores, labels)
        coefs = logistic_model.coef_.flatten()

        # Normalize coefficients
        coefs_normalized = coefs / np.sqrt(np.sum(coefs ** 2))
        coefs_dict[combination] = coefs_normalized

        # Get combination scores
        resulting_scores = scores @ coefs_normalized

        # Save resulting scores
        resulting_scores_dict[combination] = resulting_scores
        comb_list = [combination] * len(filepaths)
        df_resulting_scores = pd.DataFrame({'filepaths': filepaths, 'labels': labels, 'scores': resulting_scores, 'PC': comb_list})
        df_temp = pd.concat([df_temp, df_resulting_scores])

        # Compare current combination scores
        p = compare_scores_normal_disease(df_resulting_scores, combination, labels, save_dir)
        p_dict[combination] = p

        # Compute Vaf 
        tvaf = np.sum((coefs_normalized ** 2) * np.array(vafs))
        vafs_dict[combination] = tvaf

        # Compute AIC
        aic = compute_aic(logistic_model, scores, labels)
        aics[combination] = aic

        vafs.clear()
        scores_comb.clear()

    
    df_aics = pd.DataFrame([aics])
    df_vafs = pd.DataFrame([vafs_dict])
    df_p = pd.DataFrame([p_dict])

    return df_aics, df_vafs, df_p, df_temp, coefs_dict



def compare_scores_normal_disease(df, pc, labels, save_dir=None):
    
    all_scores = df['scores'].to_numpy()
    labels = np.array(df['labels'])

    normal = df[df['labels'] == 0]['scores']
    disease = df[df['labels'] == 1]['scores']

    if not os.path.exists(save_dir + f'plots/'):
        os.makedirs(save_dir + f'plots/')

    #Do comparisons between normal controls and diseased
    _, p = ttest_ind(normal, disease)
    plot_comparisons(normal, disease, 'Controls', 'Disease', pc, save_dir=save_dir + '/plots/')

    auc = roc_auc_score(labels, all_scores)
    RocCurveDisplay.from_predictions(labels, all_scores, name=f'PC{pc} - AUC = {auc:3f}')
    plt.title(f'PC: {pc}')


    if save_dir:
        plt.savefig(save_dir + f'plots/ROC_PC{pc}.png')

    plt.plot()
    plt.show(block=False)
    plt.close()

    return p


def compute_aic(model, X, y):
    """
    Compute the Akaike Information Criterion (AIC) for a given model.

    Args:
        model: The fitted logistic regression model
        X: The input features (principal components)
        y: The target labels (patients vs. controls)

    Returns:
        aic: The AIC value
    """
    # Number of parameters (including the intercept)
    k = X.shape[1] + 1

    # Log-likelihood of the model
    log_likelihood = -log_loss(y, model.predict_proba(X), normalize=False)

    # AIC computation
    aic = 2 * k - 2 * log_likelihood
    return aic


def normalize_vectors(vectors):
    """
    Normalize the selected vectors to have unit length.

    Args:
        vectors: 2D numpy array of selected vectors (n_voxels, n_selected_components)

    Returns:
        normalized_vectors: 2D numpy array of normalized vectors (n_voxels, n_selected_components)
    """
    norms = np.linalg.norm(vectors, axis=0)
    normalized_vectors = vectors / norms
    return normalized_vectors


def linearly_combine_vectors(normalized_gis, coefs, combination):
    # Initialize the result as a zero vector with the same length as the number of rows in gis
    result_vector = np.zeros(normalized_gis.shape[0])

    # Linearly combine the normalized vectors with the current permutation of coefficients
    for i, pc in enumerate(combination):
        result_vector += coefs[i] * normalized_gis[:, pc]


    return result_vector

def combine_pc_vectors(gis, coeffs, save_dir):
    """
    Linearly combine the principal component vectors in GIS using logistic regression coefficients.

    Args:
        df: Dataframe with the scores of normal and disease subjects to fit logistic regression
        gis: 2D numpy array of component vectors (n_voxels, n_selected_components)
        save_dir: directory where to save the results

    Returns:
        combined_pattern: 1D numpy array of the combined pattern (n_voxels,)
    """


    # Normalize the gis matrix
    normalized_gis = normalize_vectors(gis)

    # Calculate the result for each combination
    results = {}

    all_combinations = coeffs.keys()

    for combination in all_combinations:
        if len(combination) > 1:
            
            result_vector = linearly_combine_vectors(normalized_gis, coeffs.get(combination), combination)
            comb_name = "_".join(map(lambda x: str(x), combination))

            np.save(save_dir + f'vectors/GIS_vector_PC_{comb_name}.npy', result_vector)

            reshape_eigenvector_to_3d(result_vector, plot=False, save_name=f'nifti/PC_{comb_name}',
                                    save_dir=save_dir)

            results[comb_name] = result_vector

    return results


def combine_pc_vectors_roi(df, gis, roi_mask_file, save_dir):
    """
    Linearly combine the principal component vectors in GIS using logistic regression coefficients.

    Args:
        df: Dataframe with the scores of normal and disease subjects to fit logistic regression
        gis: 2D numpy array of component vectors (n_voxels, n_selected_components)
        save_dir: directory where to save the results

    Returns:
        combined_pattern: 1D numpy array of the combined pattern (n_voxels,)
    """

    pcs = df['PC'].unique()

    coeffs = dict()
    aics = dict()

    # Get the logistic regression coefficients for the individual PCs
    for pc in pcs:
        df_temp = df[df['PC'] == pc]

        all_scores = df_temp['norm_scores'].to_numpy()
        all_scores = all_scores.reshape(-1, 1)

        labels = df_temp['labels']
        coef, aic = determine_coefficients(all_scores, labels)
        coeffs[pc] = coef
        aics[pc] = aic

    # Normalize the gis matrix
    normalized_gis = normalize_vectors(gis)

    # Generate all possible combinations of the indices of coefficients
    all_combinations = []
    for r in range(2, len(coeffs) + 1):
        all_combinations.extend(combinations(range(len(coeffs)), r))

    # Calculate the result for each combination
    results = {}

    for combination in all_combinations:
        result_vector = linearly_combine_vectors(normalized_gis, coeffs, combination)
        comb_name = "_".join(map(lambda x: str(x), combination))
        #np.save(save_dir + f'/GIS_vector_PC_{comb_name}.npy', result_vector)

        reshape_eigenvector_to_roi(result_vector, roi_mask_file=roi_mask_file, plot=True, save_name=f'PC_{comb_name}',
                                  save_dir=save_dir)

        results[comb_name] = result_vector

    return results





import numpy as np
import nibabel as nb
import pandas as pd

from ssm_pca.visualization import plot_several_axes_image_array


def reshape_eigenvector_to_3d(voxel_pattern_vector, image_shape=(91, 109, 91), plot=False, save_name=None, save_dir=None):
    """
    Reshapes a specific principal component vector back into the original 3D image shape.

    Args:
        voxel_pattern_vector: 1D numpy array of voxel pattern vector corresponding to principal component index (pc) (num_voxels, )
        pc: Corresponding principal component index
        image_shape: tuple specifying the original 3D image shape (x, y, z)
        plot: Whether to display the 3D image of the pattern vector
        save_name: name of the image to save
        save_dir: directory where to save the 3D image

    Returns
        reshaped_vector: 3D numpy array of the selected component in the original image shape
    """

    reshaped_vector = voxel_pattern_vector.reshape(image_shape)

    # Display images of the GIS
    if plot:
        plot_several_axes_image_array(reshaped_vector, rotations=(90,90,90))

    if save_dir:
        nb.save(nb.Nifti1Image(reshaped_vector, np.eye(4)), save_dir + '/' + save_name)

def reshape_eigenvector_to_roi(voxel_pattern_vector, roi_mask_file, plot=False, save_name=None, save_dir=None):
    """
    Reshapes a roi principal component vector back into the original 3D roi image shape.

    Args:
        voxel_pattern_vector: 1D numpy array of voxel pattern vector corresponding to principal component index (pc) (num_rois, )
        roi_mask_file: path to the roi file
        save_name: name of the image to save
        save_dir: directory where to save 

    Returns
        reshaped_vector: 3D numpy array of the selected component in the original roi image shape
    """

    roi_mask = nb.load(roi_mask_file).get_fdata()
    reshaped_vector = np.zeros(roi_mask.shape)

    unique_labels = np.unique(roi_mask)[1:]  # Exclude background

    for region_idx, value in zip(unique_labels, voxel_pattern_vector):
        reshaped_vector[roi_mask == region_idx] = value

    if plot:
        plot_several_axes_image_array(reshaped_vector, rotations=(90,90,90), title=save_name)

  
    if save_dir:
        nb.save(nb.Nifti1Image(reshaped_vector, np.eye(4)), save_dir + '/' + save_name)
    
    return reshaped_vector

def display_gis(gis, pc, img_shape, save_dir=None):
    """Reshape the GIS vector into image format and plots it.

    Args:
        gis: GIS matrix.
        pc: Index of the GIS vector to display.
        save_dir: Directory to save the results. If None, results are not saved. Defaults to None.
    """
    save_name = f'PC_{pc}.nii'

    gis_pc = gis[:, pc]

    if save_dir:
        np.save(save_dir + f'vectors/GIS_vector_PC_{pc}.npy', gis_pc)

    reshape_eigenvector_to_3d(gis_pc, img_shape, plot=False, save_name=save_name, save_dir=save_dir + '/nifti/')

def pc_scores_to_df(score_vectors, pc, filepaths, labels):
    """Get the scores of the subjects for a given principal component.

    Args:
        score_vectors: Score vectors matrix.
        pc: Index of the principal component.
    """
    scores = score_vectors[:, pc]

    df = pd.DataFrame({'filepaths': filepaths, 'labels': labels, 'scores': scores, 'PC': pc})
    return df


from setuptools import setup, find_packages

setup(
    name='ssm_pca',
    version='0.1.0',
    author='Leonor Lopes',
    author_email='leonorserranolopes@gmail.com',
    packages=find_packages(),
    install_requires=[
        # TODO: add dependencies
    ],
    licence='MIT'
)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import scipy
from scipy.stats import ttest_ind
import nibabel as nib
from matplotlib.colors import Normalize, to_rgba



def plot_comparisons(data1, data2, label1, label2, pc, save_dir=None):
    # Perform t-test
    t_stat, p_value = ttest_ind(data1, data2)

    #plt.ion()
    # Create boxplot
    plt.figure(figsize=(10, 6))

    df = pd.DataFrame({'labels': [label1]*len(data1) + [label2]*len(data2),
                       'scores': np.concatenate([data1, data2])})

    #df = pd.DataFrame({label1: data1, label2: data2})

    sns.boxplot(data=df, x='labels', y='scores')
    sns.swarmplot(df, x='labels', y='scores', color='black', alpha=0.5)

    plt.xticks([0, 1], [label1, label2])

    # Draw line and annotate p-value
    x1, x2 = 0, 1
    #y, h, col = max(max(data1), max(data2)), 1, 'k'
    plt.text(0.5, np.max(data1), f"p = {p_value:.3f}", ha='center', va='bottom', color='k')
    plt.title(f'PC: {pc}')
    if save_dir:
        plt.savefig(save_dir + f'/box_plot_scores_PC{pc}.png')
    # Show plot
    plt.show(block=False)
    #plt.close()

def plot_slice_image_array(array, slice_idx, title=None):

    plt.ion()
    # Choose a slice to visualize (you can change this as needed)
    slice_data = array[:,:,slice_idx]

    # Display the slice using Matplotlib'notebooks imshow
    plt.imshow(slice_data, cmap='jet')
    plt.colorbar()  # Add a color bar to show intensity scale
    plt.title(title)
    plt.xlabel('X')
    plt.ylabel('Y')
    plt.show()

def plot_several_axes_image_array(array, slice_idxs:tuple=None, rotations:tuple=None, cmap=None, title:str=None, save_name:str=None):
    """
    Plot one slice per axis in a 3D image.
    Args:
        array: numpy array of the image
        slice_idxs: choose slice index to plot. Default is middle slice.
        rotations: angles to rotate the slices. Direction contrary to watch.
    """

    if len(array.shape) != 3:
        raise ValueError(f'Input image data ({title}) must be 3D')

    #plt.ion()
    # Create subplots for different views
    fig, axes = plt.subplots(1, 3, figsize=(15, 5))
    fig.suptitle(title)

    # Get min and max of array
    vmin = np.min(array)
    vmax = np.max(array)

    if slice_idxs == None:
        xy_idx = array.shape[2] // 2
        xz_idx = array.shape[1] // 2
        yz_idx = array.shape[0] // 2

        slice_idxs = [xy_idx, xz_idx, yz_idx]

    # Choose slice indices for different views
    slice_xy = array[:, :, slice_idxs[0]]
    slice_xz = array[:, slice_idxs[1], :]
    slice_yz = array[slice_idxs[2], :, :]

    # Rotate slices if needed
    if rotations:
        slice_xy = scipy.ndimage.rotate(slice_xy, rotations[0])
        slice_xz = scipy.ndimage.rotate(slice_xz, rotations[1])
        slice_yz = scipy.ndimage.rotate(slice_yz, rotations[2])


    # Plot slices for each view
    im0 = axes[0].imshow(slice_xy, cmap=cmap, vmin=vmin, vmax=vmax)
    axes[0].set_title('XY Plane')
    axes[0].set_xlabel('X')
    axes[0].set_ylabel('Y')
    axes[0].set_xticks([])
    axes[0].set_yticks([])

    axes[1].imshow(slice_xz, cmap=cmap, vmin=vmin, vmax=vmax)
    axes[1].set_title('XZ Plane')
    axes[1].set_xlabel('X')
    axes[1].set_ylabel('Z')
    axes[1].set_xticks([])
    axes[1].set_yticks([])

    axes[2].imshow(slice_yz, cmap=cmap, vmin=vmin, vmax=vmax)
    axes[2].set_title('YZ Plane')
    axes[2].set_xlabel('Y')
    axes[2].set_ylabel('Z')
    axes[2].set_xticks([])
    axes[2].set_yticks([])

    plt.colorbar(im0, cax=plt.axes([0.02, 0.12, 0.02, 0.7]), anchor=(vmin, vmax))

    if save_name:
        plt.savefig(save_name)
    plt.show(block=False)

def plot_mri_pet(mri, pet, slice_idxs:tuple=None, rotations:tuple=None, cmap='jet', title:str=None):


    if len(pet.shape) != 3:
        raise ValueError(f'Input image data ({title}) must be 3D')

    # Create subplots for different views
    fig, axes = plt.subplots(1, 3, figsize=(15, 5))
    fig.suptitle(title)

    # Get min and max of array
    vmin = np.min(pet)
    vmax = np.max(pet)

    if slice_idxs == None:
        xy_idx = pet.shape[2] // 2
        xz_idx = pet.shape[1] // 2
        yz_idx = pet.shape[0] // 2

        slice_idxs = [xy_idx, xz_idx, yz_idx]

    # Choose slice indices for different views
    slice_xy = pet[:, :, slice_idxs[0]]
    mask_xy = np.where(slice_xy == 0, 0, 0).astype(float)
    slice_xy_mri = mri[:, :, slice_idxs[0]]

    slice_xz = pet[:, slice_idxs[1], :]
    mask_xz = np.where(slice_xz == 0, 0, 0).astype(float)
    slice_xz_mri = mri[:, slice_idxs[1], :]

    slice_yz = pet[slice_idxs[2], :, :]
    mask_yz = np.where(slice_yz == 0, 0, 0).astype(float)
    slice_yz_mri = mri[slice_idxs[2], :, :]

    # Rotate slices if needed
    if rotations:
        slice_xy = scipy.ndimage.rotate(slice_xy, rotations[0])
        slice_xz = scipy.ndimage.rotate(slice_xz, rotations[1])
        slice_yz = scipy.ndimage.rotate(slice_yz, rotations[2])

        slice_xy_mri = scipy.ndimage.rotate(slice_xy_mri, rotations[0])
        slice_xz_mri = scipy.ndimage.rotate(slice_xz_mri, rotations[1])
        slice_yz_mri = scipy.ndimage.rotate(slice_yz_mri, rotations[2])


    # Plot slices for each view
    axes[0].imshow(slice_xy_mri, cmap='Greys', vmin=vmin, vmax=vmax, alpha=1)
    im0 = axes[0].imshow(slice_xy, cmap=cmap, vmin=vmin, vmax=vmax, alpha=mask_xy.astype(float))
    axes[0].set_title('XY Plane')
    axes[0].set_xlabel('X')
    axes[0].set_ylabel('Y')
    axes[0].set_xticks([])
    axes[0].set_yticks([])

    axes[1].imshow(slice_xz_mri, cmap='Greys', vmin=vmin, vmax=vmax)
    axes[1].imshow(slice_xz, cmap=cmap, vmin=vmin, vmax=vmax, alpha=mask_xz.astype(float))
    axes[1].set_title('XZ Plane')
    axes[1].set_xlabel('X')
    axes[1].set_ylabel('Z')
    axes[1].set_xticks([])
    axes[1].set_yticks([])

    axes[2].imshow(slice_yz_mri, cmap='Greys', vmin=vmin, vmax=vmax)
    axes[2].imshow(slice_yz, cmap=cmap, vmin=vmin, vmax=vmax, alpha=mask_yz.astype(float))
    axes[2].set_title('YZ Plane')
    axes[2].set_xlabel('Y')
    axes[2].set_ylabel('Z')
    axes[2].set_xticks([])
    axes[2].set_yticks([])

    plt.colorbar(im0, cax=plt.axes([0.02, 0.12, 0.02, 0.7]), anchor=(vmin, vmax))
    plt.tight_layout()
    plt.show()

def plot_2_img(pet, mri, cmap1='viridis', cmap2='viridis', norm1=None, norm2=None):

    # Create subplots for different views
    fig, axes = plt.subplots(1, 3, figsize=(15, 5))
    fig.suptitle(title)

    if slice_idxs == None:
        xy_idx = pet.shape[2] // 2
        xz_idx = pet.shape[1] // 2
        yz_idx = pet.shape[0] // 2

        slice_idxs = [xy_idx, xz_idx, yz_idx]

    # Choose slice indices for different views
    slice_xy = pet[:, :, slice_idxs[0]]
    slice_xy_mri = mri[:, :, slice_idxs[0]]

    slice_xz = pet[:, slice_idxs[1], :]
    slice_xz_mri = mri[:, slice_idxs[1], :]

    slice_yz = pet[slice_idxs[2], :, :]
    slice_yz_mri = mri[slice_idxs[2], :, :]


    norm1 = Normalize(vmin=image1.min(), vmax=image1.max())
    norm2 = Normalize(vmin=image2.min(), vmax=image2.max())

    # Create RGB combined image
    combined_image = np.zeros((*image1.shape, 4))  # RGBA format
    mask2 = image2 > 0  # Mask for image2

    # Apply colormaps
    combined_image[~mask2] = to_rgba(cmap1(norm1(image1[~mask2])))  # Image1 where image2 == 0
    combined_image[mask2] = to_rgba(cmap2(norm2(image2[mask2])))    # Image2 where image2 > 0

    # Plotting
    plt.figure(figsize=(6, 6))
    plt.title("Combined Image with Two Colormaps")
    plt.imshow(combined_image)
    plt.axis("off")
    plt.show()





import numpy as np
import nibabel as nb


def mask_img(img, mask_file=None, mask=None):
    """ Apply mask to image"""
    if mask is None:
        mask = nb.load(mask_file).get_fdata()
        
    img_masked = img * mask
    return img_masked

def create_grey_matter_mask(filelist, threshold=0.38, img_preprocess=None, to_mni_space=True, save_name=None):

    # Assuming all files in the directory are NIfTI files
    individual_masks = []

    for filename in filelist:
        img = nb.load(filename)
        array = img.get_fdata()

        if img_preprocess:
            array = img_preprocess(array)

        if to_mni_space:
            array = np.pad(array, ((6, 6), (7, 7), (11, 11)))

        max_value = np.max(array)
        mask = (array >= threshold * max_value).astype(int)
        individual_masks.append(mask)


    # Multiply all individual masks together
    multiplicative_mask = np.ones_like(individual_masks[0])

    for mask in individual_masks:
        multiplicative_mask *= mask

    # Flip the multiplicative mask left to right (mirror image)
    flipped_mask = np.flip(multiplicative_mask, axis=0)  # Assuming left-right flip along the first axis

    # Multiply the flipped mask with the original to create the final symmetrized mask
    final_mask = multiplicative_mask * flipped_mask

    if save_name:
        nb.save(nb.Nifti1Image(final_mask.astype('uint8'), img.affine), save_name)

    return final_mask


